class listNode:
    def __init__(self, size):
        self.cells = []
        self.size = size
        listSize = 2**size
        for i in range(listSize):
            self.cells.append({bin(i)[2:].zfill(size): [0, None]})
    
    def setCounter(self, cell):
        if type(cell) != str:
            cellToBinary = bin(cell)[2:].zfill(self.size)
            self.data.cells[cell][cellToBinary][0] += 1
        else:
            self.data.cells[int(cell, 2)][cell][0] += 1

    def setPointer(self, cell, pointer):
        if type(cell) != str:
            cellToBinary = bin(cell)[2:].zfill(self.size)
            self.data.cells[cell][cellToBinary][1] = pointer
        else:
            self.data.cells[int(cell, 2)][cell][1] = pointer


from .TreeNode import TreeNode
from scipy.stats import linregress
import math

def CorrelationDimension(points):
    
    try:
        dimension = getDatasetDimension(points)
    except ValueError as e:
        print(e)

    hypergridSideSize = getHyperSize(points)
    root = TreeNode(dimension)
    sidesizeValues = [math.log(hypergridSideSize, 10)]
    sumSquaredOccupancies = []
    myTree = {i: [] for i in range(dimension + 1)}

    for point in points:
        cellSize = hypergridSideSize
        bounds = [(0, cellSize)] * len(point)
        
        if(root.getCounter(0) == 0):
            root.setRootCounter()
            currentNode = root.addChild(0)
            myTree[0].append(root)
        else:
            root.setRootCounter()
            currentNode = root.getPointer(0)

        for i in range(1, dimension + 1):
            cellSize = cellSize / 2
            
            if cellSize == 0:
                break

            if math.log(cellSize, 10) not in sidesizeValues:
                sidesizeValues.append(math.log(cellSize, 10))

            cell = cellPicker(point, bounds)

            if currentNode not in myTree[i]:
                myTree[i].append(currentNode)

            if currentNode.getCounter(cell) == 0 and i != (dimension):
                currentNode.setCellCounter(cell)
                currentNode.addChild(cell)
                currentNode = currentNode.getPointer(cell)
            else:
                currentNode.setCellCounter(cell)
                currentNode = currentNode.getPointer(cell)

    for level in myTree:
        level_counter = 0
        for node in myTree[level]:
            level_counter += node.getNodeSquareCounter()
        sumSquaredOccupancies.append(math.log(level_counter, 10))

    return linregress(sidesizeValues, sumSquaredOccupancies).slope

def cellPicker(point, bounds):
    cell = []

    # Iterate over each dimension of the point
    for i, coordinate in enumerate(point):
        start, end = bounds[i]
        midpoint = (start + end) / 2  # Midpoint for the current cell (based on the current bounds)
        
        # Determine if the point is in the lower or upper half of the current bounds
        if coordinate < midpoint:
            cell.append(0)  # Inferior half
            bounds[i] = (start, midpoint)  # Update bounds to the inferior half
        else:
            cell.append(1)  # Superior half
            bounds[i] = (midpoint, end)  # Update bounds to the superior half
    
    # Return the binary string representing the cell
    return ''.join(map(str, cell))

def getDatasetDimension(dataset):
    if not dataset:
        raise ValueError("Dataset cannot be empty.")
    
    first_point_dimension = len(dataset[0])
    
    for point in dataset:
        if len(point) != first_point_dimension:
            raise ValueError("Dataset contains points with different dimensions.")
    
    return first_point_dimension

def getHyperSize(dataset):
    flattened_coords = [coordinate for point in dataset for coordinate in point]

    return max(flattened_coords)



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='fracodim',
    version='1.4',
    packages=find_packages(),
    author='Renan Gomes',
    author_email='renangcfreitas@gmail.com',
    description='Structural package of the box-counting plot method used in fractal correlation dimension estimation.',
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url='https://github.com/renangcf/fracodim',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy',
        'scipy',
    ],
)


from .listNode import listNode

nodeIndex = 0

class TreeNode(listNode):
    def __init__(self, size, index = 0):
        global nodeIndex
        self.size = size
        self.index = nodeIndex
        self.data = listNode(size)
        nodeIndex += 1

    def addChild(self, cell = 0):
        global nodeIndex
        new_child = TreeNode(self.size, nodeIndex + 1)
        self.setPointer(cell, new_child)
        
        return new_child
    
    def setCellCounter(self, cell):    
        self.setCounter(cell)

    def setRootCounter(self):
        self.setCounter(0)

    def getIndex(self):
        return self.index
    
    def getCounter(self, cell):
       if type(cell) != str:
            cellToBinary = bin(cell)[2:].zfill(self.size)
            return self.data.cells[cell][cellToBinary][0]
       else:
            return self.data.cells[int(cell, 2)][cell][0]
       
    def getPointer(self, cell):
       if type(cell) != str:
            cellToBinary = bin(cell)[2:].zfill(self.size)
            return self.data.cells[cell][cellToBinary][1]
       else:
            return self.data.cells[int(cell, 2)][cell][1]
       
    def getNodeSquareCounter(self,):
        totalSum = 0
        for cell in self.data.cells:
            for key in cell:
                totalSum += (cell[key][0]**2)
        return totalSum